#lang racket

;;;; 7. cvičenie z predmetu 1-AIN-616 Symbolické programovanie a LISP
;;;;
;;;; Meno a priezvisko:
;;;; Verzia DrRacket: 6.4
;;;; Dátum: 5.4.2016
;;;; WWW: http://www.ii.fmph.uniba.sk/cl/view.php/courses/1-AIN-616-lisp/1516ls/?lang=sk 
;;;; Kontakt: Jan.Komara(at)fmph.uniba.sk 
;;;;
;;;; Úvodná poznámka. Toto cvičenie je venované týmto témam:
;;;; - metainterpreter programovacieho jazyka LISP,
;;;; - rekurzívne nerozhodnuteľné problémy.
;;;; Zadanie sa skladá z týchto súborov: 
;;;; - Súbory maux*.rkt obsahujú definície pomocných funkcií a predikátov. 
;;;; - Súbory ex*.rkt obsahujú úlohy, ktoré máte riešiť na tomto cvičení.
;;;;
;;;; Hodnotenie.
;;;; - Počet bodov za úlohy v tomto súbore: 5
;;;; - Počet bodov za prémiove úlohy v tomto súbore: 0
;;;; - Počet bodov za projektové úlohy v tomto súbore: 0
;;;; Odovzdať do 10.4.2016 (nedeľa).

;; [Lisp]

(require racket/undefined)
(define ... undefined)
(define (query f m n) (for/list ([i (in-range m (+ n 1))]) (list i (f i))))

;; Literatúra.
;; [1] Hal Abelson and Jerry Sussman and Julie Sussman.
;;     Structure and Interpretation of Computer Programs.
;;     MIT Press, second edition, 1996. 

;; (define nil null)
;; (define apply-in-underlying-scheme apply)

;;; Kapitola. Metainterpreter programovacieho jazyka LISP.

;;; Sekcia. Vyhodnocovanie jednoduchých výrazov v prostredí väzieb.

;;; Podsekcia. Aritmetizácia.

;; Rozlišovacie položky zoznamových konštruktorov.

(define (tagged-list? exp tag)
  (and (pair? exp)
       (equal? (car exp) tag)))

;; Aritmetizácia výrazov (schéma). 

(define (make-<variant> ...) (list '<variant> ...))     ; konštruktor
(define (<variant>? exp) (tagged-list? exp '<variant>)) ; predikát
(define (<variant>-<selector> exp) ...)                 ; selektor(y)
(define (eval-<variant> exp env) ...)                   ; vyhodnocovanie

;;; Podsekcia. Vyhodnocovanie jednoduchých výrazov v prostredí väzieb.

;; Operácia my-eval(exp,env) vyhodnotí výraz exp v prostredí väzieb env.

(define (my-eval exp env)
  (cond [(<variant>? exp) (eval-<variant> exp env)]
        [else (error "Unknown expression type -- EVAL" exp)]))

;;; Podsekcia. Operácie nad prostredím väzieb.

;; Základné operácie.

(define the-empty-environment '())
(define (empty-environment? env) (null? env))
(define (enclosing-environment env) (cdr env))
(define (first-frame env) (car env))
(define the-empty-frame '())

;; Rozšírenie prostredia väzieb.

(define (make-environment frame env)
  (cons frame env))

(define (add-to-frame binding frame)
  (cons (cons (car binding) (cadr binding)) frame))

(define (extend-environment bindings env)
  (define (loop bindings frame)
    (if (null? bindings)
        (make-environment frame env)
        (loop (cdr bindings) (add-to-frame (car bindings) frame))))
  (loop bindings the-empty-frame))

;; Globálne prostredie väzieb.

(define (make-primitive-procedure proc) (list 'primitive proc))

(define frame-of-primitive-procedures
  (list
   (list 'zero? (make-primitive-procedure zero?))
   (list '+ (make-primitive-procedure +))
   (list '- (make-primitive-procedure -))
   (list '* (make-primitive-procedure *))
   ))

(define the-global-environment
  (extend-environment frame-of-primitive-procedures the-empty-environment))

;;; Podsekcia. Samo vyhodnocujúce sa výrazy.

;; Predikát.

(define (self-evaluating? exp) ...)

;; Vyhodnocovanie.

(define (eval-self-evaluating exp env) ...)

;; Testovanie.

;; (my-eval 2016 the-global-environment)
;; (my-eval "qwerty" the-global-environment)
;; (my-eval #t the-global-environment)

;;; Podsekcia. Premenné.

;; Predikát.

(define (variable? exp) ...)

;; Vyhodnocovanie.

(define (lookup-variable var env) ...)

(define (eval-variable exp env) ...)

;; Testovanie.

;; (my-eval 'x (extend-environment '([x 1]) the-global-environment))
;; (my-eval 'x (extend-environment '([x 1] [y 2]) the-global-environment))
;; (my-eval 'x (extend-environment '([y 2] [x 1]) the-global-environment))
;; (my-eval 'x (extend-environment '([x 1]) (extend-environment '([x 2]) the-global-environment)))

;;; Podsekcia. Špeciálna forma quote (výrazy s apostrofom).

;; Konštruktor.

(define (make-quoted text) (list 'quote text))

;; Predikát.

(define (quoted? exp) ...)

;; Selektor(y).

(define (text-of-quotation exp) (cadr exp))

;; Vyhodnocovanie.

(define (eval-quoted exp env) ...)

;;; Podsekcia. Špeciálna forma if (podmienkové výrazy).

;; Konštruktor.

(define (make-if predicate consequent alternative)
  (list 'if predicate consequent alternative))

;; Predikát.

(define (if? exp) ...)

;; Selektor(y).

(define (if-predicate exp) ...)
(define (if-consequent exp) ...)
(define (if-alternative exp) ...)

;; Vyhodnocovanie.

(define (eval-if exp env) ...)

;; Testovanie.

;; (my-eval '(if x y z) (extend-environment '([x #t] [y 1] [z 2]) the-global-environment))
;; (my-eval '(if x y z) (extend-environment '([x #f] [y 2] [z 1]) the-global-environment))

;;; Podsekcia. Špeciálna forma let (priradenie do lokálnej premennej).

;; Konštruktor.

(define (make-let bindings body) (list 'let bindings body))

;; Predikát.

(define (let? exp) ...)

;; Selektor(y).

(define (let-bindings exp) ...)
(define (let-body exp) ...)

;; Vyhodnocovanie.

(define (eval-bindings bindings env) ...)

(define (eval-let exp env) ...)

;; Testovanie.

;; (my-eval '(let ([x 1]) x) the-global-environment)
;; (my-eval '(let ([x 1] [y 2]) x) the-global-environment)
;; (my-eval '(let ([y 2] [x 1]) x) the-global-environment)
;; (my-eval '(let ([x 2]) (let ([x 1]) x)) the-global-environment)

;;; Podsekcia. Aplikatívne výrazy.

;; Konštruktor.

(define (make-application operator operands) (cons operator operands))

;; Predikát.

(define (application? exp) ...)

;; Selektor(y).

(define (operator exp) ...)
(define (operands exp) ...)

;; Vyhodnocovanie.

(define (eval-application exp env) ...)

;;; Podsekcia. Programy.

;; Konštruktor.

;; (define (make-primitive-procedure proc) (list 'primitive proc))

;; Predikát.

(define (primitive-procedure? proc) ...)

;; Selektor(y).

(define (primitive-procedure-implementation proc) ...)

;; Vyhodnocovanie.

(define (apply-primitive-procedure proc args) ...)

(define (my-apply procedure arguments) ...)

;; Podsekcia. Testovanie.

;; (my-eval '(+ 1 2) the-global-environment)
;; (my-eval '(let ([x 1]) (+ x 2)) the-global-environment)
;; (my-eval '(let ([x 1] [y 2]) (+ x y)) the-global-environment)
